from executor import Executor
import logging
import sys
from daemonize import daemonize
import signal
import os
from config import AgentConfig
from optparse import OptionParser
import subprocess
import tornado
import scrapydd

logger = logging.getLogger(__name__)

def init_logging(config):
    import logging.handlers
    logger = logging.getLogger()
    if not os.path.exists('logs'):
        os.mkdir('logs')
    fh = logging.handlers.TimedRotatingFileHandler('logs/scrapydd-agent.log', when='D', backupCount=7)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)

    logger.addHandler(fh)
    logger.addHandler(ch)

    if config.getboolean('debug'):
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

class Daemon():
    def __init__(self, pidfile):
        self.pidfile = pidfile
        self.subprocess_p = None
        self.pid = 0

    def start_subprocess(self):
        argv = sys.argv
        argv.remove('--daemon')
        pargs = argv
        env = os.environ.copy()
        self.subprocess_p = subprocess.Popen(pargs, env=env)
        self.subprocess_p.wait()

    def read_pidfile(self):
        try:
            with open(self.pidfile, 'r') as f:
                return int(f.readline())
        except IOError:
            return None

    def try_remove_pidfile(self):
        if os.path.exists(self.pidfile):
            os.remove(self.pidfile)

    def on_signal(self, signum, frame):
        print 'closing'
        if self.subprocess_p:
            self.subprocess_p.terminate()
        self.try_remove_pidfile()
        tornado.ioloop.IOLoop.instance().stop()

    def start(self):
        print 'Starting scrapydd agent daemon.'
        signal.signal(signal.SIGINT, self.on_signal)
        signal.signal(signal.SIGTERM, self.on_signal)
        daemonize(pidfile=self.pidfile)
        start()
        self.try_remove_pidfile()

def run(argv=None):
    if argv is None:
        argv = sys.argv
    parser = OptionParser(prog  = 'scrapydd agent')
    parser.add_option('--daemon', action='store_true', help='run scrapydd agent in daemon mode')
    parser.add_option('--pidfile', help='pid file will be created when daemon started, \
default: scrapydd-agent.pid')
    opts, args = parser.parse_args(argv)

    pidfile = opts.pidfile or 'scrapydd-agent.pid'

    if opts.daemon:
        daemon = Daemon(pidfile=pidfile)
        daemon.start()
    else:
        start()

def start():
    config = AgentConfig()
    init_logging(config)
    logging.info('------------------------')
    logging.info('Starting scrapydd agent.')
    logging.info('config %s loaded' % config._loaded_files)
    logging.info('scrapydd version : %s' % scrapydd.__version__)
    logging.info('------------------------')
    executor = Executor(config)
    executor.start()

if __name__ == '__main__':
    run()


from models import Node, Session
from tornado.ioloop import IOLoop, PeriodicCallback
import datetime
import logging
from .exceptions import *

logger = logging.getLogger(__name__)

class NodeManager():
    interval = 10
    node_timeout = 60

    def __init__(self, scheduler_manager):
        self.scheduler_manager = scheduler_manager

    def init(self):
        self.ioloop = IOLoop.current()
        self.peroid_callback = PeriodicCallback(self._poll, self.interval*1000, self.ioloop)
        self.peroid_callback.start()

    def _poll(self):
        last_heartbeat_lessthan = datetime.datetime.now() - datetime.timedelta(seconds=self.node_timeout)
        session = Session()
        for node in session.query(Node).filter(Node.last_heartbeat<last_heartbeat_lessthan, Node.isalive==1):
            logger.info('node %d expired' % node.id)
            self.scheduler_manager.on_node_expired(node.id)
            node.isalive = 0
            session.add(node)
        session.commit()
        session.close()

        logger.debug('node manager poll')

    def heartbeat(self, node_id):
        try:
            session = Session()
            node = session.query(Node).filter_by(id=node_id).first()
            if node is None:
                raise NodeExpired()
            if node.isalive == 0:
                raise NodeExpired()
            node.last_heartbeat = datetime.datetime.now()
            session.add(node)
            session.commit()
        finally:
            session.close()

    def create_node(self, remote_ip, tags=None):
        session = Session()
        node = Node()
        node.client_ip = remote_ip
        node.create_time = datetime.datetime.now()
        node.last_heartbeat = datetime.datetime.now()
        node.isalive = 1
        node.tags = tags
        session.add(node)
        session.commit()
        session.refresh(node)
        session.close()
        return node



import sys
import os
import shutil
import tempfile
from contextlib import contextmanager
from scrapy.settings import overridden_settings, Settings, iter_default_settings
from scrapy.utils.project import get_project_settings
from scrapyd.eggutils import activate_egg
from scrapyd.eggstorage import FilesystemEggStorage
import scrapyd.config
import json

@contextmanager
def project_environment(project):
    eggstorage = FilesystemEggStorage(scrapyd.config.Config())
    version, eggfile = eggstorage.get(project)
    if eggfile:
        prefix = '%s-%s-' % (project, version)
        fd, eggpath = tempfile.mkstemp(prefix=prefix, suffix='.egg')
        lf = os.fdopen(fd, 'wb')
        shutil.copyfileobj(eggfile, lf)
        lf.close()
        activate_egg(eggpath)
    else:
        eggpath = None
    try:
        assert 'scrapy.conf' not in sys.modules, "Scrapy settings already loaded"
        yield
    finally:
        if eggpath:
            os.remove(eggpath)

def main():
    if len(sys.argv) > 0:
        project = sys.argv[1]
        os.environ['SCRAPY_PROJECT'] = project
    else:
        project = os.environ['SCRAPY_PROJECT']
    with project_environment(project):
        setting = get_project_settings()
        default_settings = dict(iter_default_settings())
        setting_dict = dict(overridden_settings(setting))
        for key in setting.attributes.keys():
            if key not in default_settings:
                setting_dict[key] = setting.attributes[key].value
        print json.dumps(setting_dict)


if __name__ == '__main__':
    main()


import sys
import os, os.path
from tornado.concurrent import Future
from tornado.ioloop import IOLoop
from subprocess import Popen, PIPE
import logging
import tempfile
import scrapyd.config
from scrapyd.eggstorage import FilesystemEggStorage
import shutil
import pkg_resources
from scrapydd.exceptions import ProcessFailed, InvalidProjectEgg
import json

logger = logging.getLogger(__name__)

class ProjectWorkspace(object):
    pip = None
    python = None
    process = None
    project_workspace_dir = None
    project_check = None
    temp_dir = None

    def __init__(self, project_name):
        project_workspace_dir = os.path.abspath(os.path.join('workspace', project_name))
        self.project_workspace_dir = project_workspace_dir
        self.project_name = project_name
        self.egg_storage = FilesystemEggStorage(scrapyd.config.Config())
        if sys.platform.startswith('linux'):
            self.pip = os.path.join(project_workspace_dir, 'bin', 'pip')
            self.python = os.path.join(project_workspace_dir, 'bin', 'python')
        elif sys.platform.startswith('win'):
            self.pip = os.path.join(project_workspace_dir, 'Scripts', 'pip.exe')
            self.python = os.path.join(project_workspace_dir, 'Scripts', 'python.exe')
        else:
            raise NotImplementedError('Unsupported system %s' % sys.platform)

    def init(self):
        '''
        init project isolated workspace,
        :return: future
        '''
        future = Future()
        if os.path.exists(self.pip) and os.path.exists(self.python):
            future.set_result(self)
            return future

        logger.info('start creating virtualenv.')
        try:
            process = Popen(['virtualenv', '--system-site-packages', self.project_workspace_dir], stdout=PIPE, stderr=PIPE)
        except Exception as e:
            future.set_exception(e)
            return future
        def check_process():
            logger.debug('create virtualenv process poll.')
            retcode = process.poll()
            if retcode is not None:
                if retcode == 0:
                    future.set_result(self)
                else:
                    std_output = process.stdout.read()
                    err_output = process.stderr.read()
                    future.set_exception(ProcessFailed('Error when init workspace virtualenv ', std_output=std_output, err_output=err_output))
                return
            IOLoop.current().call_later(1, check_process)

        check_process()
        return future

    def find_project_requirements(self, project, egg_storage=None, eggf=None):
        if eggf is None:
            if egg_storage is None:
                egg_storage = FilesystemEggStorage(scrapyd.config.Config())
            version, eggf = egg_storage.get(project)
        try:
            prefix = '%s-nover-' % (project)
            fd, eggpath = tempfile.mkstemp(prefix=prefix, suffix='.egg')
            logger.debug('tmp egg file saved to %s' % eggpath)
            lf = os.fdopen(fd, 'wb')
            eggf.seek(0)
            shutil.copyfileobj(eggf, lf)
            lf.close()
            try:
                d = pkg_resources.find_distributions(eggpath).next()
            except StopIteration:
                raise ValueError("Unknown or corrupt egg")
            requirements = [str(x) for x in d.requires()]
            return requirements
        finally:
            if eggpath:
                os.remove(eggpath)

    def test_egg(self, eggf):
        future = Future()
        temp_dir = tempfile.mkdtemp('scrapydd-egg-%s' % self.project_name)
        self.temp_dir = temp_dir
        eggf.seek(0)
        egg_storage = FilesystemEggStorage(scrapyd.config.Config({'eggs_dir': os.path.join(temp_dir, 'eggs')}))
        egg_storage.put(eggf, project=self.project_name, version='1')
        eggf.seek(0)

        requirements = self._read_egg_requirements(eggf) + ['scrapyd']

        def after_spider_list(callback_future):
            logger.debug('after_spider_list')
            exc = callback_future.exception()
            if exc is not None:
                future.set_exception(exc)
                return
            spider_list = callback_future.result()
            #os.removedirs(temp_dir)
            future.set_result(spider_list)

        def after_pip_install(callback_future):
            logger.debug('after_pip_install')
            exc = callback_future.exception()
            if exc is not None:
                future.set_exception(exc)
                return

            self.spider_list(self.project_name, cwd=temp_dir).add_done_callback(after_spider_list)

        self.pip_install(requirements).add_done_callback(after_pip_install)

        return future

    def _read_egg_requirements(self, eggf):
        try:
            prefix = '%s-%s-' % (self.project_name, 0)
            fd, eggpath = tempfile.mkstemp(prefix=prefix, suffix='.egg')
            logger.debug('tmp egg file saved to %s' % eggpath)
            lf = os.fdopen(fd, 'wb')
            eggf.seek(0)
            shutil.copyfileobj(eggf, lf)
            lf.close()
            try:
                d = pkg_resources.find_distributions(eggpath).next()
            except StopIteration:
                raise ValueError("Unknown or corrupt egg")
            requirements = [str(x) for x in d.requires()]
            return requirements
        finally:
            if eggpath:
                os.remove(eggpath)


    def pip_install(self, requirements):
        logger.debug('installing requirements: %s' % requirements)
        future = Future()
        try:
            process = Popen([self.pip, 'install'] + requirements, stdout=PIPE, stderr=PIPE)
        except Exception as e:
            future.set_exception(e)
            return future
        def check_process():
            logger.debug('poll')
            retcode = process.poll()
            if retcode is not None:
                if retcode == 0:
                    future.set_result(self)
                else:
                    std_out = process.stdout.read()
                    err_out = process.stderr.read()
                    future.set_exception(ProcessFailed(std_output=std_out, err_output=err_out))
                return
            IOLoop.current().call_later(1, check_process)

        check_process()
        return future

    def spider_list(self, project, cwd=None):
        future = Future()
        try:
            env = os.environ.copy()
            env['SCRAPY_PROJECT'] = project
            process = Popen([self.python, '-m', 'scrapyd.runner', 'list'], env = env, cwd=cwd, stdout = PIPE, stderr= PIPE)
        except Exception as e:
            logger.error(e)
            future.set_exception(e)
            return future

        def check_process():
            logger.debug('poll')
            retcode = process.poll()
            if retcode is not None:
                if retcode == 0:
                    future.set_result(process.stdout.read().splitlines())
                else:
                    #future.set_exception(ProcessFailed(std_output=process.stdout.read(), err_output=process.stderr.read()))
                    future.set_exception(InvalidProjectEgg(detail=process.stderr.read()))
                return
            IOLoop.current().call_later(1, check_process)

        check_process()
        return future

    def find_project_settings(self, project):
        cwd = None
        future = Future()
        try:
            env = os.environ.copy()
            env['SCRAPY_PROJECT'] = project
            process = Popen([self.python, '-m', 'scrapydd.utils.extract_settings', project], env = env, cwd=cwd, stdout = PIPE, stderr= PIPE)
        except Exception as e:
            logger.error(e)
            future.set_exception(e)
            return future

        def check_process():
            logger.debug('find_project_settings process poll')
            retcode = process.poll()
            if retcode is not None:
                if retcode == 0:
                    proc_output = process.stdout.read()
                    logger.debug('find_project_settings output:')
                    logger.debug(proc_output)
                    future.set_result(json.loads(proc_output))
                else:
                    #future.set_exception(ProcessFailed(std_output=process.stdout.read(), err_output=process.stderr.read()))
                    future.set_exception(InvalidProjectEgg(detail=process.stderr.read()))
                return
            IOLoop.current().call_later(1, check_process)

        check_process()
        return future

    def clearup(self):
        '''
        clean up temp files.
        :return:
        '''
        if self.temp_dir and os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)

    def put_egg(self, eggfile, version):
        eggfile.seek(0)
        self.egg_storage.put(eggfile=eggfile, project=self.project_name, version=version)

    def get_egg(self, version=None):
        return self.egg_storage.get(self.project_name, version=version)

    def delete_egg(self, project, version=None):
        logger.info('deleting project eggs')
        return self.egg_storage.delete(project, version)

    def list_versions(self, project):
        return self.egg_storage.list(project)



from sqlalchemy import *
from migrate import *

meta = MetaData()

def upgrade(migrate_engine):
    meta.bind = migrate_engine
    spider_webhook = Table('spider_webhook', meta, autoload=True)
    spider_settings = Table('spider_settings', meta, autoload=True)

    for spider_webhook_row in migrate_engine.execute(spider_webhook.select()).fetchall():
        spider_id = spider_webhook_row.id
        webhook_payload = spider_webhook_row.payload_url

        migrate_engine.execute(spider_settings.insert().values(spider_id=spider_id, setting_key='webhook_payload', value=webhook_payload))


def downgrade(migrate_engine):
    # Operations to reverse the above upgrade go here.
    pass



from sqlalchemy import *
from migrate import *
from migrate.changeset.constraint import ForeignKeyConstraint

meta = MetaData()

spider_settings = Table(
    'spider_settings', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('spider_id', Integer),
    Column('setting_key', String(length=50), nullable=False),
    Column('value', String(length=200)),
)

def upgrade(migrate_engine):
    meta.bind = migrate_engine
    job_queue = Table('spider_execution_queue', meta, autoload=True)
    job_queue_slot = Column('slot', Integer, default=1)
    job_queue_slot.create(job_queue)
    spider_settings.create()

    spiders = Table('spiders', meta, autoload=True)
    spider_settings_spider_id_fk =  ForeignKeyConstraint([spider_settings.c.spider_id], [spiders.c.id])
    spider_settings_spider_id_fk.create()

def downgrade(migrate_engine):
    meta.bind = migrate_engine
    job_queue = Table('spider_execution_queue', meta, autoload=True)
    job_queue.c['slot'].drop()
    spider_settings.drop()



from sqlalchemy import *

def upgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)
    node = Table('nodes', meta, autoload=True)
    nodes_isalive_col = Column('isalive', Integer, default=1)
    nodes_isalive_col.create(node, pupulate_default=True)

def downgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)
    node = Table('nodes', meta, autoload=True)
    node.c['isalive'].drop()




from sqlalchemy import *
from migrate import *

meta = MetaData()

def upgrade(migrate_engine):
    meta.bind = migrate_engine
    job_history = Table('job_history', meta, autoload=True)
    job_history_items_count = Column('items_count', Integer)
    job_history_items_count.create(job_history)

def downgrade(migrate_engine):
    meta.bind = migrate_engine

    job_history = Table('job_history', meta, autoload=True)
    job_history.c['items_count'].drop()



from sqlalchemy import *
from migrate import *

meta = MetaData()


def upgrade(migrate_engine):
    meta.bind = migrate_engine
    webhoook_jobs = Table('webhook_jobs', meta, autoload=True)
    webhoook_jobs_spider_id = Column('spider_id', Integer)
    webhoook_jobs_spider_id.create(webhoook_jobs)

def downgrade(migrate_engine):
    webhoook_jobs = Table('webhook_jobs', meta, autoload=True)
    webhoook_jobs.c['spider_id'].drop()


from sqlalchemy import *
from migrate import *


def upgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)
    table = Table('spider_execution_queue', meta, autoload=True)
    pid_col = Column('pid', Integer)
    pid_col.create(table)


def downgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)
    table = Table('spider_execution_queue', meta, autoload=True)
    table.c['pid'].drop()



import smtplib
from email.mime.text import MIMEText


class MailSender(object):
    def __init__(self, config):
        self.smtp_server = config.get('smtp_server')
        self.smtp_port = config.getint('smtp_port')
        self.smtp_user = config.get('smtp_user')
        self.smtp_passwd = config.get('smtp_passwd')
        self.smtp_from = config.get('smtp_from')

    def send(self, to_addresses, subject, content, content_as_html=False, attchements=None):
        msg = MIMEText(content)

        msg['Subject'] = subject
        msg['From'] = self.smtp_from
        msg['To'] = to_addresses

        smtp = smtplib.SMTP(self.smtp_server, self.smtp_port)
        if self.smtp_user and self.smtp_passwd:
            smtp.login(self.smtp_user, self.smtp_passwd)
        smtp.sendmail(self.smtp_from, [to_addresses], msg.as_string())
        smtp.quit()




from sqlalchemy import *
from migrate import *

meta = MetaData()

historical_job = Table(
    'job_history', meta,
    Column('id', String(length=50), primary_key=True),
    Column('spider_id', Integer),
    Column('project_name', String(length=50)),
    Column('spider_name', String(length=50)),
    Column('fire_time', DateTime),
    Column('start_time', DateTime),
    Column('complete_time', DateTime),
    Column('status', Integer, default=0)
)

def upgrade(migrate_engine):
    meta.bind = migrate_engine
    historical_job.create()


def downgrade(migrate_engine):
    meta.bind = migrate_engine
    historical_job.drop()



from OpenSSL import crypto
from os import path, makedirs, remove
from datetime import datetime
import re
from shutil import copy

key_dir = path.join(path.dirname(path.realpath(__file__)), '..', 'keys')
key_dir = key_dir.replace('\\', '/')
index_file = key_dir + '/index.txt'


class SSLCertificateGenerator:
    key_dir = None
    index_file = None
    serial = None

    def __init__(self, key_dir=None):
        # Define key_dir
        if key_dir:
            key_dir = key_dir.replace('\\', '/')
            if not os.path.isdir(key_dir):
                raise Exception("Key Directory does not exist or is not a directory:" + key_dir)
        else:
            #key_dir = path.dirname(path.realpath(__file__)) + "/../keys"
            key_dir = path.join(path.dirname(path.realpath(__file__)), '..', 'keys')
            key_dir = key_dir.replace('\\', '/')

        self.key_dir = key_dir

        self.index_file = key_dir + '/index.txt'

        # Get serial number
        try:
            serial_file = open(key_dir + '/serial', 'r')
            self.serial = int(serial_file.readline());
            serial_file.close
        except IOError:
            self.serial = 1

    def _get_cert_dn(self, cert):
        dn = ''
        for label, value in cert.get_subject().get_components():
            dn += '/' + label + '=' + value

        return dn

    def _gen_key(self):
        # Generate new key
        key = crypto.PKey()
        key.generate_key(crypto.TYPE_RSA, 2048)
        return key

    def _create_csr(self, cert_name, key):
        req = crypto.X509Req()
        req.get_subject().CN = cert_name
        req.set_pubkey(key)
        req.sign(key, "sha256")
        return req

    def _write_key_to_file(self, key, filepath):
        key_file = open(filepath, 'w')
        key_file.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, key).decode("utf-8"))
        key_file.close()

    def _load_key_from_file(self, filepath):
        key_file = open(filepath, 'r')
        key = crypto.load_privatekey(crypto.FILETYPE_PEM, key_file.read())
        key_file.close()
        return key

    def _write_cert_to_file(self, cert, filepath):
        cert_file = open(filepath, 'w')
        cert_file.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode("utf-8"))
        cert_file.close()

    def _load_cert_from_file(self, filepath):
        cert_file = open(filepath, 'r')
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert_file.read())
        cert_file.close()
        return cert

    def _write_csr_to_file(self, csr, filepath):
        csr_file = open(filepath, 'w')
        csr_file.write(crypto.dump_certificate_request(crypto.FILETYPE_PEM, csr).decode("utf-8"))
        csr_file.close()

    def _load_csr_from_file(self, filepath):
        csr_file = open(filepath, 'r')
        csr = crypto.load_certificate_request(crypto.FILETYPE_PEM, csr_file.read())
        csr_file.close()
        return csr

    def _write_pfx_to_file(self, pkcs12, filepath):
        pkcs12_file = open(filepath, 'wb')
        pkcs12_file.write(pkcs12.export())
        pkcs12_file.close()

    def _write_crl_to_file(self, crl, ca_cert, ca_key, filepath):
        # Write CRL file
        crl_file = open(filepath, 'w')
        crl_file.write(crl.export(ca_cert, ca_key, days=365).decode("utf-8"))
        crl_file.close()

    def _load_crl_from_file(self, filepath):
        try:
            crl_file = open(filepath, 'r')
            crl = crypto.load_crl(crypto.FILETYPE_PEM, crl_file.read())
            crl_file.close()
        except IOError:
            # Create new CRL file if it doesn't exist
            crl = crypto.CRL()

        return crl

    def _sign_csr(self, req, ca_key, ca_cert, cert_org=False, cert_ou=False, usage=3, days=3650, alt_names=[]):
        expiry_seconds = days * 86400

        # Create and sign certificate
        cert = crypto.X509()
        cert.set_version(2)
        cert.set_subject(req.get_subject())
        if cert_org:
            cert.get_subject().O = cert_org
        else:
            cert.get_subject().O = ca_cert.get_subject().O
        if cert_ou:
            cert.get_subject().OU = cert_ou
        else:
            cert.get_subject().OU = ca_cert.get_subject().OU
        cert.set_serial_number(self.serial)
        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(expiry_seconds)
        cert.set_issuer(ca_cert.get_subject())
        cert.set_pubkey(req.get_pubkey())
        if usage == 1:
            cert.add_extensions([
                crypto.X509Extension(b"basicConstraints", True, b"CA:TRUE, pathlen:0"),
                crypto.X509Extension(b"keyUsage", True, b"keyCertSign, cRLSign"),
                crypto.X509Extension(b"subjectKeyIdentifier", False, b"hash", subject=cert)
            ])
        elif usage == 2:
            cert.add_extensions([
                crypto.X509Extension(b"extendedKeyUsage", True, b"serverAuth"),
            ])
        elif usage == 3:
            cert.add_extensions([
                crypto.X509Extension(b"extendedKeyUsage", True, b"clientAuth"),
            ])

        # Add alt names
        if alt_names:
            for name in alt_names:
                name = "DNS:" + name
            cert.add_extensions([
                crypto.X509Extension(b"subjectAltName", False, b"DNS:" + ",DNS:".join(alt_names).encode("utf-8"))
            ])

        cert.sign(ca_key, "sha256")

        # Write to index.txt
        db_line = "V\t" + cert.get_notBefore().decode("utf-8") + "\t\t" + hex(
            int(cert.get_serial_number())) + "\tunknown\t" + str(cert.get_subject())[18:-2] + "\n"
        index_file = open(key_dir + '/index.txt', 'a')
        index_file.write(db_line)
        index_file.close()

        # Write updated serial file
        serial_file = open(key_dir + '/serial', 'w')
        serial_file.write(str(self.serial + 1))
        serial_file.close()

        return cert

    def gen_ca(self, cert_org="Thinkbox Software", cert_ou="IT", days=3650):
        expiry_seconds = days * 86400

        # Generate key
        key = crypto.PKey()
        key.generate_key(crypto.TYPE_RSA, 2048)

        # Set up and sign CA certificate
        ca = crypto.X509()
        ca.set_version(2)
        ca.set_serial_number(1)
        ca.get_subject().CN = "CA"
        ca.get_subject().O = cert_org
        ca.get_subject().OU = cert_ou
        ca.gmtime_adj_notBefore(0)
        ca.gmtime_adj_notAfter(expiry_seconds)
        ca.set_issuer(ca.get_subject())
        ca.set_pubkey(key)
        ca.add_extensions([
            crypto.X509Extension(b"basicConstraints", True, b"CA:TRUE, pathlen:0"),
            crypto.X509Extension(b"keyUsage", True, b"keyCertSign, cRLSign"),
            crypto.X509Extension(b"subjectKeyIdentifier", False, b"hash", subject=ca)
        ])
        ca.sign(key, "sha256")

        # Create key directory if it doesn't exist
        if not path.exists(key_dir):
            makedirs(key_dir)

        # Write CA certificate to file
        self._write_cert_to_file(ca, self.key_dir + '/ca.crt')

        # Write CA key to file
        self._write_key_to_file(key, self.key_dir + '/ca.key')

    def get_ca_key(self):
        self._load_key_from_file(path.join(self.key_dir, 'ca.key'))

    def get_ca_cert(self):
        self._load_cert_from_file(path.join(self.key_dir, 'ca.crt'))

    def gen_cert(self, cert_name, cert_org=False, cert_ou=False, usage=3, days=3650, alt_names=[]):
        # usage: 1=ca, 2=server, 3=client
        if cert_name == "":
            raise Exception("Certificate name cannot be blank")

        # Load CA certificate
        ca_cert = self._load_cert_from_file(self.key_dir + '/ca.crt')

        # Load CA key
        ca_key = self._load_key_from_file(self.key_dir + '/ca.key')

        # Generate new key
        key = self._gen_key()

        # Create CSR
        req = self._create_csr(cert_name, key)

        # Sign CSR
        cert = self._sign_csr(req, ca_key, ca_cert, cert_org=cert_org, cert_ou=cert_ou, usage=usage, days=days,
                              alt_names=alt_names)

        # Write new key file
        self._write_key_to_file(key, self.key_dir + '/' + cert_name + '.key')

        # Write new certificate file
        self._write_cert_to_file(cert, self.key_dir + '/' + cert_name + '.crt')

    def gen_pfx(self, cert_name):
        if cert_name == "":
            raise Exception("Certificate name cannot be blank")

        # Load CA certificate
        ca_cert = self._load_cert_from_file(self.key_dir + '/ca.crt')

        # Load Certificate
        cert = self._load_cert_from_file(self.key_dir + '/' + cert_name + '.crt')

        # Load Private Key
        key = self._load_key_from_file(self.key_dir + '/' + cert_name + '.key')

        # Set up PKCS12 structure
        pkcs12 = crypto.PKCS12()
        pkcs12.set_ca_certificates([ca_cert])
        pkcs12.set_certificate(cert)
        pkcs12.set_privatekey(key)

        # Write PFX file
        self._write_pfx_to_file(pkcs12, self.key_dir + '/' + cert_name + '.pfx')

    def gen_csr(self, name, out_dir):
        key = self._gen_key()
        csr = self._create_csr(name, key)
        self._write_key_to_file(key, out_dir + '/' + name + '.key')
        self._write_csr_to_file(csr, out_dir + '/' + name + '.csr')

    def sign_csr(self, csr_path):
        csr = self._load_csr_from_file(csr_path)
        ca_key = self._load_key_from_file(key_dir + '/ca.key')
        ca_cert = self._load_cert_from_file(key_dir + '/ca.crt')
        cert = self._sign_csr(csr, ca_key, ca_cert)
        self._write_cert_to_file(cert, self.key_dir + '/' + csr.get_subject().CN + '.crt')

    def revoke_cert(self, cert_name):
        # Load CA certificate
        ca_cert = self._load_cert_from_file(self.key_dir + '/ca.crt')

        # Load CA Key
        ca_key = self._load_key_from_file(self.key_dir + '/ca.key')

        # Load Certificate
        cert = self._load_cert_from_file(self.key_dir + '/' + cert_name + '.crt')

        # Load Private Key
        key = self._load_key_from_file(self.key_dir + '/' + cert_name + '.key')

        # Load CRL File
        crl = self._load_crl_from_file(self.key_dir + '/crl.pem')

        print ('Revoking ' + cert_name + ' (Serial: ' + str(cert.get_serial_number()) + ')')

        # Revoke certificate
        revoked = crypto.Revoked()
        revoked.set_serial(hex(int(cert.get_serial_number()))[2:].encode("utf-8"))
        revoked.set_reason(b'unspecified')
        revoked.set_rev_date(datetime.utcnow().strftime('%Y%m%d%H%M%SZ').encode("utf-8"))
        crl.add_revoked(revoked)

        # Write CRL file
        self._write_crl_to_file(crl, ca_cert, ca_key, key_dir + '/crl.pem')

        # Update index file
        index_file = open(key_dir + '/index.txt', 'r')
        index_file_new = open(key_dir + '/index.txt.new', 'w')

        for line in index_file.readlines():
            line_split = re.split('\t', line)
            if int(line_split[3], 16) == cert.get_serial_number():
                new_line = 'R\t' + line_split[1] + '\t' + revoked.get_rev_date().decode("utf-8") + '\t' + line_split[
                    3] + '\t' + line_split[4] + '\t' + line_split[5]
                index_file_new.write(new_line)
            else:
                index_file_new.write(line)

        index_file.close()
        index_file_new.close()

        copy('keys/index.txt.new', 'keys/index.txt')
        remove('keys/index.txt.new')

    def renew_crl(self):
        # Load CA certificate
        ca_cert = self._load_cert_from_file(self.key_dir + '/ca.crt')

        # Load CA key
        ca_key = self._load_key_from_file(self.key_dir + '/ca.key')

        # Load CRL File
        crl = self._load_crl_from_file(self.key_dir + '/crl.pem')

        # Write CRL file
        self._write_crl_to_file(crl, ca_cert, ca_key, key_dir + '/crl.pem')


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='SSL Certificate Generator')

    arg_group = parser.add_mutually_exclusive_group()
    arg_group.add_argument('--ca', action='store_true', help='Generate a CA certificate')
    arg_group.add_argument('--intermediate-ca', action='store_true', help='Generate an intermediate ca certificate')
    arg_group.add_argument('--server', action='store_true', help='Generate a server certificate')
    arg_group.add_argument('--client', action='store_true', help='Generate a client certificate')
    arg_group.add_argument('--pfx', action='store_true', help='Generate a PFX File')
    arg_group.add_argument('--revoke', action='store_true', help='Revoke a certificate')
    arg_group.add_argument('--renew-crl', action='store_true', help='Renew CRL')

    parser.add_argument('--cert-name', help='Certificate name (required with --server, --client, and --pfx)')
    parser.add_argument('--cert-org', help='Certificate organization (required with --ca)')
    parser.add_argument('--cert-ou', help='Certificate organizational unit (required with --ca)')
    parser.add_argument('--alt-name', help='Subject Alternative Name', action='append')

    args = parser.parse_args()

    sslgen = SSLCertificateGenerator()

    if args.ca:
        error = False
        if args.cert_name:
            print('Error: Certificate name was specified.  CA certificate is always named "ca"')
            error = True
        if not args.cert_ou:
            print("Error: No OU specified")
            error = True
        if not args.cert_org:
            print("Error: No organization specified")
            error = True
        if error:
            exit(1)

        sslgen.gen_ca(cert_org=args.cert_org, cert_ou=args.cert_ou)

    elif args.intermediate_ca:
        if not args.cert_name:
            print("Error: No certificate name specified")
            exit(1)

        sslgen.gen_cert(args.cert_name, cert_org=args.cert_org, cert_ou=args.cert_ou, usage=1)

    elif args.server:
        if not args.cert_name:
            print("Error: No certificate name specified")
            exit(1)

        sslgen.gen_cert(args.cert_name, cert_org=args.cert_org, cert_ou=args.cert_ou, usage=2, alt_names=args.alt_name)

    elif args.client:
        if not args.cert_name:
            print("Error: No certificate name specified")
            exit(1)

        sslgen.gen_cert(args.cert_name, cert_org=args.cert_org, cert_ou=args.cert_ou, usage=3, alt_names=args.alt_name)

    elif args.pfx:
        if not args.cert_name:
            print("Error: No certificate name specified")
            exit(1)

        sslgen.gen_pfx(args.cert_name)

    elif args.revoke:
        if not args.cert_name:
            print("Error: No certificate name specified")
            exit(1)

        sslgen.revoke_cert(args.cert_name)

    elif args.renew_crl:
        sslgen.renew_crl()

    else:
        print("Error: Certificate type must be specified using [--ca|--server|--client|--pfx]")
        exit(1)


from sqlalchemy import *
from migrate import *

meta = MetaData()

def upgrade(migrate_engine):
    meta.bind = migrate_engine
    spider_execution_queue = Table('spider_execution_queue', meta, autoload=True)
    spider_execution_queue_tag = Column('tag', String(length=50))
    spider_execution_queue_tag.create(spider_execution_queue)


def downgrade(migrate_engine):
    meta.bind = migrate_engine
    spider_execution_queue = Table('spider_execution_queue', meta, autoload=True)
    spider_execution_queue.c['tag'].drop()



from sqlalchemy import create_engine, schema, Column, desc
from sqlalchemy.types import Integer, String, DateTime
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship
from migrate.versioning.api import version_control, upgrade
from migrate.exceptions import DatabaseAlreadyControlledError
import os
from contextlib import contextmanager
from scrapydd.config import Config

metadata = schema.MetaData()
Base = declarative_base(metadata=metadata)

#database_url = 'mysql://scrapydd:scrapydd@localhost/scrapydd'
database_url = 'sqlite:///database.db'
engine = create_engine(database_url)
Session = sessionmaker(bind=engine, expire_on_commit=False)
_Session = None


class Project(Base):
    __tablename__ = 'projects'

    id = Column(Integer, primary_key=True)
    name = Column(String(length=50))
    version = Column(String(length=50))

class Spider(Base):
    __tablename__ = 'spiders'

    id = Column(Integer, primary_key=True)
    project_id = Column(Integer, ForeignKey('projects.id'))
    project = relationship('Project')
    name = Column(String(length=50))

Project.spiders = relationship("Spider", order_by = Spider.id)

class Trigger(Base):
    __tablename__ = 'triggers'

    id = Column(Integer, primary_key=True)
    spider_id = Column(Integer, ForeignKey('spiders.id'))
    cron_pattern = Column(String(length=50))

Spider.triggers = relationship("Trigger", order_by=Trigger.id)


class JobStatus(Base):
    __tablename__ = 'job_status'
    id = Column(Integer, primary_key=True)
    name = Column(String(length=50), nullable=False)
    desc = Column(String(length=200))


class SpiderExecutionQueue(Base):
    __tablename__ = 'spider_execution_queue'

    id = Column(String(length=50), primary_key=True)
    spider_id = Column(Integer)
    slot = Column(Integer, default=1)
    project_name = Column(String(length=50))
    spider_name = Column(String(length=50))
    fire_time = Column(DateTime)
    start_time = Column(DateTime)
    node_id = Column(Integer)
    status = Column(Integer, ForeignKey('job_status.id'), default=0)
    status_obj = relationship('JobStatus')
    update_time = Column(DateTime)
    pid = Column(Integer)
    tag = Column(String(length=50))

class Node(Base):
    __tablename__ = 'nodes'

    id = Column(Integer, primary_key=True)
    client_ip = Column(String(length=50))
    create_time = Column(DateTime)
    last_heartbeat = Column(DateTime)
    isalive = Column(Integer)
    tags = Column(String(length=200))


class HistoricalJob(Base):
    __tablename__ = 'job_history'

    id = Column(String(length=50), primary_key=True)
    spider_id = Column(Integer, ForeignKey('spiders.id'))
    spider = relationship('Spider')
    project_name = Column(String(length=50))
    spider_name = Column(String(length=50))
    fire_time = Column(DateTime)
    start_time = Column(DateTime)
    complete_time = Column(DateTime)
    status = Column(Integer, ForeignKey('job_status.id'), default=0)
    status_obj = relationship('JobStatus')
    log_file = Column(String(500))
    items_file = Column(String(500))
    items_count = Column(Integer)

Spider.historical_jobs = relationship("HistoricalJob", order_by=desc(HistoricalJob.start_time))


class SpiderWebhook(Base):
    __tablename__ = 'spider_webhook'

    id = Column(String(length=50), primary_key=True)
    payload_url = Column(String(length=250))


class WebhookJob(Base):
    __tablename__ = 'webhook_jobs'

    id = Column(Integer, primary_key=True, autoincrement=True)
    job_id = Column(String(length=50))
    spider_id = Column(Integer, ForeignKey('spiders.id'))
    payload_url = Column(String(length=50))
    items_file = Column(String(length=250))
    status = Column(Integer, ForeignKey('job_status.id'), default=0)
    status_obj = relationship('JobStatus')


class SpiderSettings(Base):
    __tablename__ = 'spider_settings'

    id = Column(Integer, primary_key=True, autoincrement=True)
    spider_id = Column(Integer, ForeignKey('spiders.id'))
    setting_key = Column(String(length=50), nullable=False)
    value = Column(String(length=200))


class SpiderParameter(Base):
    __tablename__ = 'spider_parameters'

    id = Column(Integer, primary_key=True, autoincrement=True)
    spider_id = Column(Integer, ForeignKey('spiders.id'))
    parameter_key = Column(String(length=50), nullable=False)
    value = Column(String(length=200))

Spider.parameters = relationship('SpiderParameter', order_by=SpiderParameter.parameter_key)


def init_database(config=None):
    if config is None:
        config = Config()

    global database_url
    global engine
    global _Session
    database_url = config.get('database_url')
    engine = create_engine(database_url)
    _Session = sessionmaker(bind=engine, expire_on_commit=False)
    #global Session
    db_repository = os.path.join(os.path.dirname(__file__), 'migrates')
    try:
        version_control(url=database_url, repository=db_repository)
    except DatabaseAlreadyControlledError:
        pass
    upgrade(database_url, db_repository)

def _make_session():
    global _Session
    return _Session()

Session = _make_session



@contextmanager
def session_scope():
    """Provide a transactional scope around a series of operations."""
    session = _make_session()
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()



#!/usr/bin/env python3
"""Post data streamer for tornadoweb 4.0"""
import os
import re
import random
import tempfile

class SizeLimitError(Exception):
    pass

class PostDataStreamer:
    """Parse a stream of multpart/form-data.

    Useful for request handlers decorated with tornado.web.stream_request_body"""
    SEP = b"\r\n"
    LSEP = len(SEP)
    PAT_HEADERVALUE = re.compile(r"""([^:]+):\s+([^\s;]+)(.*)""")
    PAT_HEADERPARAMS = re.compile(r""";\s*([^=]+)=\"(.*?)\"(.*)""")

    # Encoding for the header values. Only header name and parameters
    # will be decoded. Streamed data will remain binary.
    # This is required because multipart/form-data headers cannot
    # be parsed without a valid encoding.
    header_encoding = "UTF-8"

    def __init__(self, total, tmpdir=None):
        self.buf = b""
        self.dlen = None
        self.delimiter = None
        self.in_data = False
        self.headers = []
        self.parts = []
        self.total = total
        self.received = 0
        self.tmpdir = tmpdir

    def _get_raw_header(self,data):
        idx = data.find(self.SEP)
        if idx>=0:
            return (data[:idx], data[idx+self.LSEP:])
        else:
            return (None, data)

    def receive(self, chunk):
        self.received += len(chunk)
        self.on_progress()
        self.buf += chunk

        if not self.delimiter:
            self.delimiter, self.buf = self._get_raw_header(self.buf)
            if self.delimiter:
                self.delimiter+=self.SEP
                self.dlen = len(self.delimiter)
            elif len(self.buf)>1000:
                raise Exception("Cannot find multipart delimiter")
            else:
                return

        while True:
            if self.in_data:
                if (len(self.buf)>3*self.dlen):
                    idx = self.buf.find(self.SEP+self.delimiter)
                    if idx>=0:
                        self.feed_part(self.buf[:idx])
                        self.end_part()
                        self.buf = self.buf[idx+len(self.SEP+self.delimiter):]
                        self.in_data = False
                    else:
                        limit = len(self.buf)-2*self.dlen
                        self.feed_part(self.buf[:limit])
                        self.buf = self.buf[limit:]
                        return
                else:
                    return
            if not self.in_data:
                while True:
                    header, self.buf = self._get_raw_header(self.buf)
                    if header==b"":
                        assert(self.delimiter)
                        self.in_data = True
                        self.begin_part(self.headers)
                        self.headers = []
                        break
                    elif header:
                        self.headers.append(self.parse_header(header))
                    else:
                        # Header is None, not enough data yet
                        return

    def parse_header(self,header):
        header = header.decode(self.header_encoding)
        res = self.PAT_HEADERVALUE.match(header)
        if res:
            name,value,tail = res.groups()
            params = {}
            hdr = {"name":name,"value":value,"params":params}
            while True:
                res = self.PAT_HEADERPARAMS.match(tail)
                if not res:
                    break
                fname,fvalue,tail = res.groups()
                params[fname] = fvalue
            return hdr
        else:
            return {"value":header}

    def begin_part(self,headers):
        """Internal method called when a new part is started."""
        self.fout = tempfile.NamedTemporaryFile(
            dir=self.tmpdir, delete=False)
        self.part = {
            "headers":headers,
            "size":0,
            "tmpfile":self.fout
        }
        self.parts.append(self.part)

    def feed_part(self,data):
        """Internal method called when content is added to the current part."""
        self.fout.write(data)
        self.part["size"] += len(data)

    def end_part(self):
        """Internal method called when receiving the current part has finished."""
        # Will not close the file here, so we will be able to read later.
        #self.fout.close()            
        #self.fout.flush() This is not needed because we update part["size"]
        self.fout.flush()
        pass

    def finish_receive(self):
        """Call this after the last receive() call.

        You MUST call this before using the parts."""
        if self.in_data:
            idx = self.buf.rfind(self.SEP+self.delimiter[:-2])
            if idx>0:
                self.feed_part(self.buf[:idx])
            self.end_part()

    def release_parts(self):
        """Call this to remove the temporary files."""
        for part in self.parts:
            part["tmpfile"].close()
            os.unlink(part["tmpfile"].name)

    def get_part_payload(self, part):
        """Return the contents of a part.

        Warning: do not use this for big files!"""
        fsource = part["tmpfile"]
        fsource.seek(0)
        return fsource.read()

    def get_part_ct_params(self, part):
        """Get content-disposition parameters.

        If there is no content-disposition header then it returns an
        empty list."""
        for header in part["headers"]:
            if header.get("name","").lower().strip()=="content-disposition":
                return header.get("params",[])
        return []

    def get_part_ct_param(self, part, pname, defval=None):
        """Get parameter for a part.

        @param part: The part
        @param pname: Name of the parameter, case insensitive
        @param defval: Value to return when not found.
        """
        ct_params = self.get_part_ct_params(part)
        for name in ct_params:
            if name.lower().strip()==pname:
                return ct_params[name]
        return defval

    def get_part_name(self, part):
        """Get name of a part.

        When not given, returns None."""
        return self.get_part_ct_param(part, "name", None)

    def get_parts_by_name(self, pname):
        """Get a parts by name.

        @param pname: Name of the part. This is case sensitive!

        Attention! A form may have posted multiple values for the same
        name. So the return value of this method is a list of parts!"""
        res = []
        for part in self.parts:
            name = self.get_part_name(part)
            if name==pname:
                res.append(part)
        return res

    def get_values(self, fnames, size_limit=100*1024):
        """Return a dictionary of values for the given field names.

        @param fnames: A list of field names.
        @param size_limit: Maximum size of the value of a single field.
            If a field's size exceeds this then SizeLimitError is raised.

        Warning: do not use this for big file values.
        Warning: a form may have posted multiple values for a field name.
            This method returns the first available value for that name.
            To get all values, use the get_parts_by_name method.
        Tip: use get_nonfile_names() to get a list of field names
            that are not originally files.
        """
        res = {}
        for fname in fnames:
            parts = self.get_parts_by_name(fname)
            if not parts:
                raise KeyError("No such field: %s"%fname)
            size = parts[0]["size"]
            if size>size_limit:
                raise SizeLimitError("Part size=%s > limit=%s"%(size, size_limit))
            res[fname] = self.get_part_payload(parts[0])
        return res

    def get_nonfile_names(self):
        """Get a list of part names are originally not files.

        It examines the filename attribute of the content-disposition header.
        Be aware that these fields still may be huge in size."""
        res = []
        for part in self.parts:
            filename = self.get_part_ct_param(part, "filename", None)
            if filename is None:
                name = self.get_part_name(part)
                if name:
                    res.append(name)
        return res

    def examine(self):
        """Debugging method for examining received data."""
        print("============= structure =============")
        for idx,part in enumerate(self.parts):
            print("PART #",idx)
            print("    HEADERS")
            for header in part["headers"]:
                print("        ",repr(header.get("name","")),"=",repr(header.get("value","")))
                params = header.get("params",None)
                if params:
                    for pname in params:
                        print("            ",repr(pname),"=",repr(params[pname]))
            print("    DATA")
            print("        SIZE", part["size"])
            print("        LOCATION",part["tmpfile"].name)
            if part["size"]<80:
                print("        PAYLOAD:",repr(self.get_part_payload(part)))
            else:
                print("        PAYLOAD:","<too long...>")
        print("========== non-file values ==========")
        print(self.get_values(self.get_nonfile_names()))


    def on_progress(self):
        """Override this function to handle progress of receiving data."""
        pass # Received <self.received> of <self.total>


class MultipartRequestBodyProducer():
    def __init__(self, datagen):
        self.datagen = datagen

    def __call__(self, write):
        for part in self.datagen:
            write(part)
